#pragma once

#include <pcl\point_types.h>
#include <pcl\point_cloud.h>
#include <pcl\segmentation\sac_segmentation.h>
#include "data_def_api.h"
#include "data_def.h"
#include "logger.h"
#include "imstorage_manager.h"
#include "grab_occlusion.h"

namespace graft_cv {
	class CRootStockGrabPoint {
	public:
		CRootStockGrabPoint(ConfigParam&c, CGcvLogger*pLog = 0);
		~CRootStockGrabPoint();
		int grab_point_detect(PositionInfo& posinfo);
		float* get_raw_point_cloud(int &data_size);
		int load_data(float*pPoint, int pixel_size, int pt_size, int dtype, const char* fn = 0);
		void set_image_saver(CImStoreManager** ppis) { m_ppImgSaver = ppis; }
	private:
		//global configure parameters
		ConfigParam& m_cparam;
		CGcvLogger * m_pLogger;
		int m_dtype;
		string m_pcdId;

		//ʷֲλϢ
		CStemResultInfos* m_pStemInfos;
		std::string m_stem_info_file;
		std::vector<CStemResult> m_root_centers;  //ͨm_pStemInfosȡʷλ
		std::vector<bool> m_root_center_with_seedling;	//m_root_centersλǷֲ

		//ڼ¼һzֵб12ŵ
		float m_1st_row_zmean_rs = -1.0;
		float m_1st_row_zmean_sc = -1.0;

		CImStoreManager** m_ppImgSaver;

		//ͼƬڵ
		ImgInfo* m_pImginfoResult;//ʶͼƬ
		

		pcl::PointCloud<pcl::PointXYZ>::Ptr m_raw_cloud;
		double m_cloud_mean_dist;	


		void clear_imginfo();

		int read_ply_file(const char* fn);
		double compute_nearest_neighbor_distance(pcl::PointCloud<pcl::PointXYZ>::Ptr);		

		/////////////////////////////////////////////////////////////////////////////////////
		//Ҷ޳,ͨŷʽ࣬ȻжÿһǷҶ
		//   ֵ⣺ҶӺ;ͨҶʱ
		void leaf_filter(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud, 
			std::vector<int> &stem_cloud_idx
			);
		
		void leaf_filter_ror(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,	//input 
			std::vector<int> &stem_cloud_idx,				//output, ˺ĵ
			std::vector<int>& leaf_idx						//output, Ҷӵĵ
		);
		void cloud_mean_dist(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,	//input 
			double& mean_dist
		);
		//////////////////////////////////////////////////////////////////////////////////////
		void  CRootStockGrabPoint::find_fork(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_line_cloud,
			float& max_dist,//
			int& max_idx //߽ĵ
		);


		void gen_all_seedling_positions(
			pcl::PointXYZ&key_center,	//룬֪
			std::vector<pcl::PointXYZ>&candidate_center	//б
		);
		bool find_seedling_position_key(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,
			std::vector<int> &first_seedling_cloud_idx,
			pcl::PointXYZ&xz_center,
			pcl::ModelCoefficients::Ptr& lmodel,
			int& first_row_size //صһֲ
		);
		// С
		void nms_box(
			std::vector<pcl::PointXYZ>&clt_root_v,	//ĿĿλ
			std::vector<float>&valid_box_cc_dist,	//Ŀĵĵľ룬ԽԽ
			float hole_step_radius,					//Ŀ뾶
			std::vector<pcl::PointXYZ>& target_toot // ֵ
		);
		//ָͨλ,ȡֵƷǷľ澥λñ浽target_filtered
		void line_filter(pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,  //
			float radius,												//룬ȡ뾶
			float ymin,													//룬yСֵ
			float ymax,													//룬yֵ
			std::vector<pcl::PointXYZ>&target_root,						//룬λ
			std::vector<pcl::PointXYZ>&target_filtered,					//λ
			std::vector<pcl::PointXYZ>&target_filtered_root,			//ϸ
			std::vector<std::vector<int>>&target_filtered_element,		//ϵindex
			std::vector<pcl::ModelCoefficients::Ptr>& target_filtered_models//ֱģ
		);
		void get_line_project_hist(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_line_cloud,	//input	ֱߵ
			pcl::ModelCoefficients::Ptr line_model,				//input		
			std::vector<int>& count_hist					    //  Чֱ1mmڵ
		);

		/////////////////////////////////////////////////
		void find_seedling_position(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,
			std::vector<int> &first_seedling_cloud_idx,
			pcl::PointXYZ&xz_center
		);
		void crop_nn_analysis(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,
			pcl::PointCloud<pcl::PointXYZ>::Ptr seed_cloud,
			double dist_mean,
			std::vector<double>&mass_indices,
			std::vector<int>& idx
		);

		void euclidean_clustering_ttsas(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,
			double d1, double d2,
			std::vector<pcl::PointXYZ>&cluster_center,
			std::vector<std::vector<int>> &clustr_member
		);
		void cal_obb_2d(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,
			int axis,
			double &dx_obb,
			double &dy_obb,
			double &angle_obb);

		
        //֪λãҳпλ
		void tilted_seedling_discover(
			std::vector<pcl::PointXYZ>&cluster_center,	//룬֪
			std::vector<pcl::PointXYZ>&tilted_center	//б
		);
		
		//ͨȽֱߵƺԭʼͬλxzķΧȷ˵Ƿ޸ŵľ
		void get_optimal_seed_compare(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,		//input
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_line_cloud,	//input
			pcl::ModelCoefficients::Ptr line_model,			//input
			pcl::PointXYZ&pt,								//
			pcl::PointXYZ &pt_ref,							//	ص㾥ڵĲο			
			float& stem_width_mu,								//,ȣֱ
			float& stem_deflection							//,Ӷȣƫֱĵľ룬
		);
		//㾥ȣƵֱ߾95λ
		double calculate_deflection(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_line_cloud,	//input
			pcl::ModelCoefficients::Ptr line_model			//input
		);
		//һĿռ䷶Χҳֱߣ辥ֱ߷ֲģֱϵĵindex
		void find_line_in_cube(
			pcl::PointCloud<pcl::PointXYZ>::Ptr,	//룬
			pcl::PointXYZ& min_pt_ex,				//룬
			pcl::PointXYZ& max_pt_ex,				//룬
			std::vector<int>& out_idx				//ֱϵindex 
			);

		//ɽͼƬ
		void gen_result_img(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud_raw,//룬δ˵in_cloud_raw, 
			pcl::PointCloud<pcl::PointXYZ>::Ptr,	//룬cloud_dowm_sampled, 
			pcl::PointXYZ& selected_pt,				//룬selected_pt, 
			pcl::PointXYZ& selected_pt_ref,        //룬selected_pt_ref,
			const PositionInfo& posinfo,					//룬
			cv::Mat& rst_img						//ͼƬ 640*1280
		);
		//ʷλãӦλõжϴλǷ
		void occluded_seedling_detect_by_leaf(
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,	//input 			
			std::vector<int>& leaf_idx);	//input
		int get_point_count_inbox(const CStemResult& sr, //input
			pcl::PointXYZ& aabb_min,	//output
			pcl::PointXYZ& aabb_max,	//output
			pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud,	//input 			
			std::vector<int>& leaf_idx);
		//ûм⵽
		void no_seedling_detected_post_process(
			int first_row_seedling_number,		//input
			int& selected_idx,					//output, ѡroot_center
			pcl::PointXYZ& selected_pt,			//output
			pcl::PointXYZ& selected_pt_ref,		//output
			PositionInfo& posinfo				//output
		);
		//⵽
		void had_seedling_detected_post_process(
			int first_row_seedling_number,		//input
			float stem_width_mu,				//input
			float stem_deflection,				//input
			int& selected_idx,					//output, ѡroot_center
			pcl::PointXYZ& selected_pt,			//input-output
			pcl::PointXYZ& selected_pt_ref,		//input-output
			PositionInfo& posinfo				//output
		);

		void viewer_cloud(pcl::PointCloud<pcl::PointXYZ>::Ptr, std::string&winname);
		void viewer_cloud(pcl::PointCloud<pcl::PointXYZRGB>::Ptr, std::string&winname);

		void viewer_cloud_debug(
			pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud, 
			pcl::PointXYZ&p,//ץȡ
			pcl::PointXYZ &p_ref,//ֲ
			pcl::PointXYZ &root_pt,
			std::string&winname);

		void viewer_cloud_cluster(pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud, std::vector<pcl::PointXYZ>cluster_center, std::string&winname);
		void viewer_cloud_cluster_box(
			pcl::PointCloud<pcl::PointXYZ>::Ptr cloud,
			std::vector<pcl::PointXYZ>&cluster_center,
			std::vector<pcl::PointXYZ>&cluster_box,
			std::vector<std::vector<int> >& clt_line_idx,
			std::string&winname);
	};

};